using System;
using System.Collections.Generic;
using System.Threading;
using Dashboard.ImmediateDraw;
using Dashboard.Controls;
using Dashboard.Media;
using Dashboard.PAL;
using Dashboard.Typography;

namespace Dashboard
{
    /// <summary>
    /// Main class for Dashboard applications.
    /// </summary>
    public class DbApplication
    {
        /// <summary>
        /// The application platform driver.
        /// </summary>
        public IDashboardPlatform Platform { get; }

        /// <summary>
        /// Title of the application.
        /// </summary>
        public string? Title
        {
            get => Platform.Title;
            set => Platform.Title = value;
        }

        /// <summary>
        /// Application icon.
        /// </summary>
        public QImage? Icon
        {
            get => Platform.Icon;
            set => Platform.Icon = value;
        }

        public PAL.Dash? MainPort { get; private set; } = null;

        public FontProvider FontProvider { get; }

        /// <summary>
        /// List of media loaders, drivers that load media such as images and fonts.
        /// </summary>
        public List<MediaLoader> MediaLoaders { get; } = new List<MediaLoader>();

        public DbApplication(IDashboardPlatform platform)
        {
            Platform = platform;
            FontProvider = new FontProvider(this);
            Current = this;
        }

        public IDisposable? GetMedia(object key, MediaHint hint)
        {
            IDisposable? disposable = null;

            foreach (MediaLoader loader in MediaLoaders)
            {
                disposable = loader.GetMedia(key, hint);

                if (disposable != null)
                    break;
            }

            return disposable;
        }

        public IDisposable? GetMedia<T>(T key, MediaHint hint)
        {
            IDisposable? disposable = null;

            foreach (MediaLoader loader in MediaLoaders)
            {
                if (loader is MediaLoader<T> typedLoader)
                {
                    disposable = typedLoader.GetMedia(key, hint);

                    if (disposable != null)
                        break;
                }
            }

            return disposable;
        }

        private DrawList cmd = new DrawList();

        public void Run(View mainView, bool yield = true)
        {
            Init(mainView);

            while (RunSync())
            {
                if (yield)
                {
                    Thread.Yield();
                }
            }
        }

        public void Init(View mainView)
        {
            MainPort = new PAL.Dash(Platform) { UIElement = mainView };
            MainPort.EventRaised += (sender, ea) => mainView.NotifyEvent(sender, ea);
        }

        public bool RunSync()
        {
            if (!MainPort!.IsValid)
                return false;

            Platform.ProcessEvents(false);

            if (MainPort.IsValid)
            {
                cmd.Clear();
                MainPort.Paint(cmd);
            }

            return true;
        }

        public static DbApplication Current { get; private set; } = null!;

        public static void SetCurrentApplication(DbApplication application)
        {
            Current = application;
        }
    }
}