using Dashboard.VertexGenerator;
using Dashboard.Media;
using OpenTK.Graphics.OpenGL4;
using System;
using System.IO;
using System.Collections.Generic;
using System.Linq;
using OpenTK.Mathematics;

namespace Dashboard.OpenGL
{
    public class GL21Driver : IDisposable
    {
        private int program;
        private int v2Position;
        private int fZIndex;
        private int v2TexPos;
        private int fTexLayer;
        private int v4Color;
        private int m4Transforms;
        private int fMaxZ;
        private int iEnableSdf;
        private int iEnableTexture;
        private int iAlphaDiscard;
        private int fSdfThreshold;
        private int tx2d;
        private int tx2darray;

        private bool isDiposed;
        private readonly Dictionary<DrawCallQueue, DrawData> data = new Dictionary<DrawCallQueue, DrawData>();
        private readonly TextureManager textures = new TextureManager();

        public bool IsInit { get; private set; } = false;
        public event Action<GL21Driver>? OnGCDispose;

        public GL21Driver()
        {
        }

        ~GL21Driver()
        {
            Dispose(false);
        }

        public void Init()
        {
            if (IsInit) return;

            int vs = CreateShader(ShaderType.VertexShader, "Dashboard.res.gl21.vert");
            int fs = CreateShader(ShaderType.FragmentShader, "Dashboard.res.gl21.frag");

            program = GL.CreateProgram();
            GL.AttachShader(program, vs);
            GL.AttachShader(program, fs);

            GL.LinkProgram(program);

            if (CheckProgram(program, out string msg) == false)
            {
                GraphicsException ex = new GraphicsException("Could not link shader program.");
                ex.Data.Add("Program Info Log", msg);
            }

            GL.DeleteShader(vs);
            GL.DeleteShader(fs);

            v2Position = GL.GetAttribLocation(program, nameof(v2Position));
            fZIndex    = GL.GetAttribLocation(program, nameof(fZIndex));
            v2TexPos   = GL.GetAttribLocation(program, nameof(v2TexPos));
            fTexLayer  = GL.GetAttribLocation(program, nameof(fTexLayer));
            v4Color    = GL.GetAttribLocation(program, nameof(v4Color));

            m4Transforms   = GL.GetUniformLocation(program, nameof(m4Transforms));
            fMaxZ          = GL.GetUniformLocation(program, nameof(fMaxZ));
            fSdfThreshold  = GL.GetUniformLocation(program, nameof(fSdfThreshold));
            iEnableSdf     = GL.GetUniformLocation(program, nameof(iEnableSdf));
            iEnableTexture = GL.GetUniformLocation(program, nameof(iEnableTexture));
            iAlphaDiscard  = GL.GetUniformLocation(program, nameof(iAlphaDiscard));
            tx2d           = GL.GetUniformLocation(program, nameof(tx2d));
            tx2darray      = GL.GetUniformLocation(program, nameof(tx2darray));

            IsInit = true;
        }

        private void AssertInit()
        {
            if (!IsInit) throw new InvalidOperationException("Initialize the driver first.");
        }

        public void Draw(DrawCallQueue queue, in QRectangle view)
        {
            AssertInit();

            if (!data.TryGetValue(queue, out DrawData? draw))
            {
                draw = new DrawData(this, queue);
                data.Add(queue, draw);
            }

            // This already binds the vertex array for me.
            draw.PrepareFrame();


            Vector2 size = view.Size;
            QMat4.Orthographic(out QMat4 viewMatrix, view);

            GL.Viewport(0, 0, (int)view.Size.X, (int)view.Size.Y);
            GL.UseProgram(program);
            GL.Uniform1(fMaxZ, (float)(queue.ZDepth+1));
            GL.Uniform1(fSdfThreshold, 0.5f);
            GL.Uniform1(tx2d, 0);
            GL.Enable(EnableCap.Blend);
            GL.BlendFunc(BlendingFactor.SrcAlpha, BlendingFactor.OneMinusSrcAlpha);
            GL.Enable(EnableCap.ScissorTest);
            GL.Enable(EnableCap.DepthTest);
            GL.DepthFunc(DepthFunction.Less);

            foreach (DrawCall call in queue)
            {
                GL.Scissor(
                    (int)MathF.Round(call.Bounds.Min.X),
                    (int)MathF.Round(size.Y - call.Bounds.Max.Y),
                    (int)MathF.Round(call.Bounds.Size.X),
                    (int)MathF.Round(call.Bounds.Size.Y));
                QMat4.Translation(out QMat4 modelMatrix, call.Bounds.Min.X, call.Bounds.Min.Y, 0);
                QMat4 modelView = viewMatrix * modelMatrix;
                GL.UniformMatrix4(m4Transforms, 1, false, ref modelView.M11);

                GL.ActiveTexture(TextureUnit.Texture0);
                GL.BindTexture(TextureTarget.Texture2D, 0);
                if (call.Texture != null)
                {
                    GL.Uniform1(iEnableSdf, call.Texture.IsSdf ? 1 : 0);
                    GL.Uniform1(iAlphaDiscard, 1);

                    if (call.Texture.Depth > 1)
                    {
                        GL.Uniform1(iEnableTexture, 3);
                        GL.ActiveTexture(TextureUnit.Texture1);
                        GL.BindTexture(TextureTarget.Texture2D, textures.GetTexture(call.Texture));
                    }
                    else
                    {
                        GL.Uniform1(iEnableTexture, 2);
                        GL.ActiveTexture(TextureUnit.Texture0);
                        GL.BindTexture(TextureTarget.Texture2D, textures.GetTexture(call.Texture));
                    }
                }
                else
                {
                    GL.Uniform1(iEnableTexture, 0);
                }

                GL.DrawElements(PrimitiveType.Triangles, call.Count, DrawElementsType.UnsignedInt, sizeof(int)*call.Start);
            }

            GL.Disable(EnableCap.ScissorTest);
            GL.Disable(EnableCap.DepthTest);
            GL.Disable(EnableCap.Blend);
        }

        public void ClearDrawQueue(DrawCallQueue queue)
        {
            AssertInit();

            if (!data.TryGetValue(queue, out DrawData? draw))
                return;
            draw.Dispose();
            data.Remove(queue);
        }

        private static int CreateShader(ShaderType type, string name)
        {
            StreamReader source = new StreamReader(typeof(GL21Driver).Assembly.GetManifestResourceStream(name) ?? throw new Exception("Resource not found."));
            string text = source.ReadToEnd();
            source.Dispose();

            int shader = GL.CreateShader(type);
            GL.ShaderSource(shader, text);
            GL.CompileShader(shader);

            if (CheckShader(shader, out string msg) == false)
            {
                GraphicsException ex = new GraphicsException($"Failed to compile {type} shader stage.");
                ex.Data.Add("Shader Stage", type);
                ex.Data.Add("Shader Info Log", msg);
                ex.Data.Add("Shader Source", text);
                throw ex;
            }

            return shader;
        }

        private static bool CheckShader(int shader, out string message)
        {
            message = string.Empty;

            GL.GetShader(shader, ShaderParameter.CompileStatus, out int i);

            if (i == 0)
            {
                message = GL.GetShaderInfoLog(shader);
                return false;
            }

            return true;
        }

        private static bool CheckProgram(int program, out string message)
        {
            message = string.Empty;

            GL.GetProgram(program, GetProgramParameterName.LinkStatus, out int i);

            if (i == 0)
            {
                message = GL.GetProgramInfoLog(program);

                return false;
            }

            return true;
        }

        private void Dispose(bool disposing)
        {
            if (isDiposed) return;

            if (!IsInit)
            {
                isDiposed = true;
                return;
            }

            if (!disposing)
            {
                if (OnGCDispose == null)
                {
                    throw new Exception("This object must strictly be disposed from the owning thread, not GC");
                }
                else
                {
                    OnGCDispose(this);
                    return;
                }
            }

            GL.DeleteProgram(program);

            foreach (DrawData datum in data.Values)
            {
                datum.Dispose();
            }

            isDiposed = true;
            GC.SuppressFinalize(this);
        }

        public void Dispose() => Dispose(true);

        private class DrawData : IDisposable
        {
            public DrawCallQueue Queue { get; }
            public int VertexArray { get; }

            private readonly GL21Driver driver;
            private int vbo1, vbo2;
            private int ebo1, ebo2;

            public DrawData(GL21Driver driver, DrawCallQueue queue)
            {
                Queue = queue;
                this.driver = driver;
                VertexArray = GL.GenVertexArray();
                GL.GenBuffers(1, out vbo1);
                GL.GenBuffers(1, out vbo2);
                GL.GenBuffers(1, out ebo1);
                GL.GenBuffers(1, out ebo2);
                isDisposed = false;
            }

            public void PrepareFrame()
            {
                int vbo, ebo;
                vbo = Swap(ref vbo1, ref vbo2);
                ebo = Swap(ref ebo1, ref ebo2);

                if (Queue.VertexCount == 0 || Queue.ElementCount == 0)
                    return;

                GL.BindVertexArray(VertexArray);
                GL.BindBuffer(BufferTarget.ArrayBuffer, vbo);
                GL.BufferData(BufferTarget.ArrayBuffer, DbVertex.Stride * Queue.VertexCount, Queue.VertexArray, BufferUsageHint.StreamDraw);

                GL.VertexAttribPointer(driver.v2Position, 2, VertexAttribPointerType.Float, false, DbVertex.Stride, DbVertex.PositionOffset);
                GL.VertexAttribPointer(driver.fZIndex, 1, VertexAttribPointerType.UnsignedInt, false, DbVertex.Stride, DbVertex.ZIndexOffset);
                GL.VertexAttribPointer(driver.v2TexPos, 2, VertexAttribPointerType.Float, false, DbVertex.Stride, DbVertex.TextureCoordinatesOffset);
                GL.VertexAttribPointer(driver.fTexLayer, 1, VertexAttribPointerType.Float, false, DbVertex.Stride, DbVertex.TextureLayerOffset);
                GL.VertexAttribPointer(driver.v4Color, 4, VertexAttribPointerType.UnsignedByte, true, DbVertex.Stride, DbVertex.ColorOffset);
                GL.EnableVertexAttribArray(driver.v2Position);
                GL.EnableVertexAttribArray(driver.fZIndex);
                GL.EnableVertexAttribArray(driver.v2TexPos);
                GL.EnableVertexAttribArray(driver.v4Color);
                GL.EnableVertexAttribArray(driver.fTexLayer);

                GL.BindBuffer(BufferTarget.ElementArrayBuffer, ebo);
                GL.BufferData(BufferTarget.ElementArrayBuffer, Queue.ElementCount * sizeof(int), Queue.ElementArray, BufferUsageHint.StreamDraw);

                int Swap(ref int a, ref int b)
                {
                    a ^= b;
                    b ^= a;
                    a ^= b;
                    return a;
                }
            }

            private bool isDisposed;
            public void Dispose()
            {
                if (isDisposed) return;

                GL.DeleteVertexArray(VertexArray);
                GL.DeleteBuffer(vbo1);
                GL.DeleteBuffer(vbo2);
                GL.DeleteBuffer(ebo1);
                GL.DeleteBuffer(ebo2);
            }
        }
    }

    internal class TextureManager : IDisposable
    {
        private readonly Dictionary<QImage, int> textures = new Dictionary<QImage, int>();
        private readonly HashSet<QImage> imagesNotUsed = new HashSet<QImage>();
        private bool isDisposed = false;

        public void BeginFrame()
        {
            if (imagesNotUsed.Count > 0)
            {
                foreach (QImage image in imagesNotUsed)
                {
                    GL.DeleteTexture(textures[image]);
                }

                imagesNotUsed.Clear();
            }

            foreach (QImage image in textures.Keys)
            {
                imagesNotUsed.Add(image);
            }
        }

        public int GetTexture(QImage image)
        {
            if (textures.TryGetValue(image, out int texture))
            {
                return texture;
            }

            if (image.Depth > 1)
            {
                texture = UploadTexture3d(image);
            }
            else
            {
                texture = UploadTexture2d(image);
            }
           
            return textures[image] = texture;
        }

        public int UploadTexture3d(QImage image3d)
        {
            int texture = GL.GenTexture();

            GL.BindTexture(TextureTarget.Texture2DArray, texture);

            image3d.LockBits3d(out QImageLock lck, QImageLockOptions.Default);
                GL.TexImage3D(TextureTarget.Texture2DArray, 0, PixelInternalFormat.Rgba, lck.Width, lck.Height, lck.Depth, 0, s_InternalFormat[lck.Format], s_PixelType[lck.Format], lck.ImagePtr);
            image3d.UnlockBits();

            GL.TexParameter(TextureTarget.Texture2DArray, TextureParameterName.TextureMinFilter, (int)TextureMinFilter.Linear);
            GL.TexParameter(TextureTarget.Texture2DArray, TextureParameterName.TextureMagFilter, (int)TextureMinFilter.Linear);

            return texture;
        }

        public int UploadTexture2d(QImage image2d)
        {
            int texture = GL.GenTexture();

            GL.BindTexture(TextureTarget.Texture2D, texture);

            image2d.LockBits2d(out QImageLock lck, QImageLockOptions.Default);
            GL.TexImage2D(TextureTarget.Texture2D, 0, PixelInternalFormat.Rgba, lck.Width, lck.Height, 0, s_InternalFormat[lck.Format], s_PixelType[lck.Format], lck.ImagePtr);
            image2d.UnlockBits();

            GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureMinFilter, (int)TextureMinFilter.Linear);
            GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureMagFilter, (int)TextureMinFilter.Linear);

            switch (image2d.InternalFormat)
            {
            case QImageFormat.RedU8:
            case QImageFormat.RedF:
                GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureSwizzleR, (int)All.Red);
                GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureSwizzleG, (int)All.Red);
                GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureSwizzleB, (int)All.Red);
                GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureSwizzleA, (int)All.One);
                break;
            case QImageFormat.AlphaU8:
            case QImageFormat.AlphaF:
                GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureSwizzleR, (int)All.One);
                GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureSwizzleG, (int)All.One);
                GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureSwizzleB, (int)All.One);
                GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureSwizzleA, (int)All.Alpha);
                break;
            }

            return texture;
        }

        public void Dispose()
        {
            if (isDisposed)
                return;

            isDisposed = true;

            int[] ids = textures.Values.ToArray();
            GL.DeleteTextures(ids.Length, ref ids[0]);
        }

        private static readonly Dictionary<QImageFormat, PixelFormat> s_InternalFormat = new Dictionary<QImageFormat, PixelFormat>()
        {
            [QImageFormat.AlphaF]   = PixelFormat.Alpha,
            [QImageFormat.AlphaU8]  = PixelFormat.Alpha,
            [QImageFormat.RedF]     = PixelFormat.Red,
            [QImageFormat.RedU8]    = PixelFormat.Red,
            [QImageFormat.RgbF]     = PixelFormat.Rgb,
            [QImageFormat.RgbU8]    = PixelFormat.Rgb,
            [QImageFormat.RgbaU8]   = PixelFormat.Rgba,
            [QImageFormat.RgbaF]    = PixelFormat.Rgba,
        };

        private static readonly Dictionary<QImageFormat, PixelType> s_PixelType = new Dictionary<QImageFormat, PixelType>()
        {
            [QImageFormat.AlphaF]   = PixelType.Float,
            [QImageFormat.RedF]     = PixelType.Float,
            [QImageFormat.RgbF]     = PixelType.Float,
            [QImageFormat.RgbaF]    = PixelType.Float,
            [QImageFormat.AlphaU8]  = PixelType.UnsignedByte,
            [QImageFormat.RedU8]    = PixelType.UnsignedByte,
            [QImageFormat.RgbU8]    = PixelType.UnsignedByte,
            [QImageFormat.RgbaU8]   = PixelType.UnsignedByte,
        };
    }
}